// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "DawnHostVita.hpp"
#include "game/DawnGame.hpp"

using namespace Dawn;

void vitaTestCube() {
  float colors[] = {1.0, 1.0, 1.0, 1.0, 0.0, 0.0, 1.0, 0.0, 1.0, 1.0, 1.0, 1.0}; // Colors for a face
  float vertices_front[] = {-0.5f, -0.5f, -0.5f, 0.5f, -0.5f, -0.5f, -0.5f, 0.5f, -0.5f, 0.5f, 0.5f, -0.5f}; // Front Face
  float vertices_back[] = {-0.5f, -0.5f, 0.5f, 0.5f, -0.5f, 0.5f, -0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f}; // Back Face
  float vertices_left[] = {-0.5f, -0.5f, -0.5f, -0.5f, 0.5f, -0.5f, -0.5f, -0.5f, 0.5f, -0.5f, 0.5f, 0.5f}; // Left Face
  float vertices_right[] = {0.5f, -0.5f, -0.5f, 0.5f, 0.5f, -0.5f, 0.5f, -0.5f, 0.5f, 0.5f, 0.5f, 0.5f}; // Right Face
  float vertices_top[] = {-0.5f, -0.5f, -0.5f, 0.5f, -0.5f, -0.5f, -0.5f, -0.5f, 0.5f, 0.5f, -0.5f, 0.5f}; // Top Face
  float vertices_bottom[] = {-0.5f, 0.5f, -0.5f, 0.5f, 0.5f, -0.5f, -0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f}; // Bottom Face

  uint16_t indices[] = {
    0, 1, 2, 1, 2, 3, // Front
    4, 5, 6, 5, 6, 7, // Back
    8, 9,10, 9,10,11, // Left
    12,13,14,13,14,15, // Right
    16,17,18,17,18,19, // Top
    20,21,22,21,22,23  // Bottom
  };

  // Creating colors array for all faces
  float color_array[12*6];
  int i;
  for (i=0;i<12*6;i++){
  color_array[i] = colors[i % 12];
  }

  // Creating vertex array for all faces
  float vertex_array[12*6];
  memcpy(&vertex_array[12*0], &vertices_front[0], sizeof(float) * 12);
  memcpy(&vertex_array[12*1], &vertices_back[0], sizeof(float) * 12);
  memcpy(&vertex_array[12*2], &vertices_left[0], sizeof(float) * 12);
  memcpy(&vertex_array[12*3], &vertices_right[0], sizeof(float) * 12);
  memcpy(&vertex_array[12*4], &vertices_top[0], sizeof(float) * 12);
  memcpy(&vertex_array[12*5], &vertices_bottom[0], sizeof(float) * 12);

  // Setting clear color
  glClearColor (0.0f, 0.0f, 0.0f, 0.0f);

  // Initializing mvp matrix with a perspective full screen matrix
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(90.0f, 960.f/544.0f, 0.01f, 100.0f);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glTranslatef(0.0f, 0.0f, -3.0f); // Centering the cube

  // Enabling depth test
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LESS);

  // Main loop
  for (;;){
  // Clear color and depth buffers
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // Drawing our cube with vertex arrays
  glEnableClientState(GL_VERTEX_ARRAY);
  glEnableClientState(GL_COLOR_ARRAY);
  glVertexPointer(3, GL_FLOAT, 0, vertex_array);
  glColorPointer(3, GL_FLOAT, 0, color_array);
  glRotatef(1.0f, 0.0f, 0.0f, 1.0f); // Rotating cube at each frame by 1 on axis x and axis w
  glRotatef(0.5f, 0.0f, 1.0f, 0.0f); // Rotating cube at each frame by 0.5 on axis x and 1.0 on axis z
  glDrawElements(GL_TRIANGLES, 6*6, GL_UNSIGNED_SHORT, indices);
  glDisableClientState(GL_VERTEX_ARRAY);
  glDisableClientState(GL_COLOR_ARRAY);

  // Performing buffer swap
  vglSwapBuffers(GL_FALSE);
  }
}


DawnHost::DawnHost() {

}

int32_t DawnHost::init(DawnGame *game) {
  this->game = game;

  vglInit(0x800000);
  vglWaitVblankStart(GL_TRUE);

  // Override the defaults
  game->renderManager.backBuffer.setSize(DAWN_VITA_WIDTH, DAWN_VITA_HEIGHT);

  // Initialize the game
  auto result = game->init();
  if(result != DAWN_GAME_INIT_RESULT_SUCCESS) return result;

  // Hard-Load the first scene assets.
  if(game->scene != nullptr) {
    auto assets = game->scene->getRequiredAssets();
    game->assetManager.queueLoad(assets);
    game->assetManager.syncLoad();
    game->scene->stage();
  }

  return DAWN_HOST_INIT_RESULT_SUCCESS;
}

int32_t DawnHost::start(DawnGame *game) {
  double_t time, newTime;
  float_t fDelta;
  int32_t updateResult;

  // Main Render Loop 
  time = 0.0f;
	for(;;) {
    // Determine the delta.
    float_t fDelta = 1.0f / 60.0f;
    time = newTime;

    // Perform update
    updateResult = this->update(game, fDelta);

    // Did the update complete successfully?
    if(updateResult == DAWN_HOST_UPDATE_RESULT_EXIT) {
      break;
    } else if(updateResult != DAWN_HOST_UPDATE_RESULT_SUCCESS) {
      return updateResult;
    }
    
		// Performing buffer swap
		vglSwapBuffers(GL_FALSE);
  }

  return DAWN_HOST_START_RESULT_EXIT_SUCCESS;
}

int32_t DawnHost::update(DawnGame *game, float_t delta) {
  // Tick game.
  auto ret = game->update(delta);
  switch(ret) {
    case DAWN_GAME_UPDATE_RESULT_SUCCESS:
      return DAWN_HOST_UPDATE_RESULT_SUCCESS;
    
    case DAWN_GAME_UPDATE_RESULT_EXIT:
      return DAWN_HOST_UPDATE_RESULT_EXIT;

    default:
      return ret;
  }
}

void DawnHost::unload(DawnGame *game) {
  vglEnd();
}

DawnHost::~DawnHost() {
}


int main(int argc, char **args) {
  int32_t result;

  // Create the host
  auto host = new DawnHost();
  auto game = new DawnGame(host);

  // Initialize the host and error check
  result = host->init(game);
  switch(result) {
    case DAWN_HOST_INIT_RESULT_SUCCESS:
      break;
    default:
      return result;
  }

  // Request the main loop to start running.
  result = host->start(game);
  switch(result) {
    case DAWN_HOST_START_RESULT_SUCCESS:
      break;
    case DAWN_HOST_START_RESULT_EXIT_SUCCESS:
      break;
    default:
      return result;
  }

  // Main loop finished without errors, cleanup
  host->unload(game);

  delete game;
  delete host;

  // Success
  return 0;
}